//---------------------------------------------------------------------------------------------------------------------------------------------------
// <copyright file="ShaderParameters.cs" company="DarkWynter Studios">
//     Copyright (C)2007 DarkWynter Studios.  All rights reserved.
// </copyright>
//---------------------------------------------------------------------------------------------------------------------------------------------------
// {License Information: Creative Commons}
//---------------------------------------------------------------------------------------------------------------------------------------------------

namespace ElementalGame
{
    #region Using Statements
    using System;
    using System.Collections.Generic;
    using Microsoft.Xna.Framework;
    using Microsoft.Xna.Framework.Content;
    using Microsoft.Xna.Framework.Graphics;
    #endregion

    public class GpuProcessor
    {
        // Primary Render target
        Viewport IndexedViewport;
        Viewport SpatialViewport;

        Vec4 locationMap;               // Spacial Data Structure
        int LOCATION_MAP_SCALE = 32;    // Scale our 128x128 black texture to fit the number of objects

        // Quad Vertex Buffer
        private VertexBuffer quad_vertexBuffer;                 // Buffer to store vertices
        private VertexPositionNormalTexture[] quad_vertices;    // Array to store vertex Position, Normal and Texture information
        private VertexDeclaration quad_vertexDeclaration;       // Vertex Declaration

        // Point Vertex Buffer
        private VertexBuffer point_vertexBuffer;                // Buffer to store vertices
        private VertexPositionNormalTexture[] point_vertices;   // Array to store vertex Position, Normal and Texture information
        private VertexDeclaration point_vertexDeclaration;      // Vertex Declaration

        // Shader Handles
        Effect effect;                          // The program handle
        EffectParameter DT;                     // Time difference between frames
        EffectParameter IndexedPass;            // Keeps track of which Physics pass we are on
        Effect eleMentalEffect;                 // The handle to ElementalGPU.fx
        EffectParameter currentPositionParam;   // To save the CurrentPosition texture to ElementalGPU
        EffectParameter scaleRange;             // To save the Scale Range to ElementalGPU
        EffectParameter minimumValue;           // To save the Minimum Value to ElementalGPU
        bool initializeConstants = true;

        public static int INDEXED_TARGET_SIZE = 256;
        public static int SPATIAL_TARGET_SIZE = 512;

        int SPATIAL_VERTEX_SIZE = 0;

        public GpuProcessor()
        {
            // Set up the Indexed viewport
            IndexedViewport.X = 0;
            IndexedViewport.Y = 0;
            IndexedViewport.Width = INDEXED_TARGET_SIZE;
            IndexedViewport.Height = INDEXED_TARGET_SIZE;
            IndexedViewport.MinDepth = ElementalGame.graphics.GraphicsDevice.Viewport.MinDepth;
            IndexedViewport.MaxDepth = ElementalGame.graphics.GraphicsDevice.Viewport.MaxDepth;

            // Set up the Spatial viewport
            SpatialViewport.X = 0;
            SpatialViewport.Y = 0;
            SpatialViewport.Width = SPATIAL_TARGET_SIZE;
            SpatialViewport.Height = SPATIAL_TARGET_SIZE;
            SpatialViewport.MinDepth = ElementalGame.graphics.GraphicsDevice.Viewport.MinDepth;
            SpatialViewport.MaxDepth = ElementalGame.graphics.GraphicsDevice.Viewport.MaxDepth;

            // Set up Uniform shader connections
            effect = ElementalGame.content.Load<Effect>("PhysicsGPU/GpuCollisionPhysics");
            IndexedPass = effect.Parameters["IndexedPass"];     
            DT = effect.Parameters["dt"];

            eleMentalEffect = ElementalGame.content.Load<Effect>("Shaders/ElementalGPU");
            try
            {
                currentPositionParam = eleMentalEffect.Parameters["CurrentPosition"];
                scaleRange = eleMentalEffect.Parameters["scaleRange"];
                minimumValue = eleMentalEffect.Parameters["minimumValue"];
            }
            catch
            { throw new Exception("shit happened"); }
            // Create Location Map
            locationMap = new Vec4(effect, "LocationMap", "locationMap", true);

            // Create geometry used in render passes
            CreateQuad();
            CreatePoints();
        }

        private void CreateQuad()
        {
            GraphicsDevice gd = ElementalGame.graphics.GraphicsDevice;

            // Create a 4 point vertex buffer
            quad_vertices = new VertexPositionNormalTexture[4];
            quad_vertexDeclaration = new VertexDeclaration(gd, VertexPositionNormalTexture.VertexElements);
            quad_vertexBuffer = new VertexBuffer(gd, VertexPositionNormalTexture.SizeInBytes * 4,
                                            ResourceUsage.None, ResourceManagementMode.Automatic);

            // Create quad with width and height = 1
            quad_vertices[0] = new VertexPositionNormalTexture(new Vector3(-0.5f, -0.5f, 0), 
                                                               new Vector3(-0.5f, -0.5f, 1), 
                                                               new Vector2(0, 1.0f));           // 3rd quadrent
            quad_vertices[1] = new VertexPositionNormalTexture(new Vector3(0.5f, -0.5f, 0), 
                                                               new Vector3(0.5f, -0.5f, 1), 
                                                               new Vector2(1.0f, 1.0f));        // 4th quadrent
            quad_vertices[2] = new VertexPositionNormalTexture(new Vector3(0.5f, 0.5f, 0), 
                                                               new Vector3(0.5f, 0.5f, 1), 
                                                               new Vector2(1.0f, 0));           // 1st quadrent
            quad_vertices[3] = new VertexPositionNormalTexture(new Vector3(-0.5f, 0.5f, 0), 
                                                               new Vector3(-0.0f, 1.0f, 1), 
                                                               new Vector2(0, 0));              // 2nd quadrent

            // Set the data
            quad_vertexBuffer.SetData<VertexPositionNormalTexture>(quad_vertices);
        }
        
        private void CreatePoints()
        {
            GraphicsDevice gd = ElementalGame.graphics.GraphicsDevice;

            point_vertices = new VertexPositionNormalTexture[SPATIAL_TARGET_SIZE * SPATIAL_TARGET_SIZE];
            point_vertexDeclaration = new VertexDeclaration(gd, VertexPositionNormalTexture.VertexElements);
            point_vertexBuffer = new VertexBuffer(gd, VertexPositionNormalTexture.SizeInBytes * SPATIAL_TARGET_SIZE * SPATIAL_TARGET_SIZE,
                ResourceUsage.None, ResourceManagementMode.Automatic);

            // Row by row, create a point for each pixel between 0-1
            int index = 0;
            float pixelLength = SPATIAL_TARGET_SIZE - 1.0f;
            for (int j = 0; j <= pixelLength; j++)
            {
                for (int i = 0; i <= pixelLength; i++)
                {
                    Vector3 pPos = new Vector3(((float)i / pixelLength), ((float)j / pixelLength), 0.0f);
                    Vector3 nPos = new Vector3(((float)i / pixelLength), ((float)j / pixelLength), 1.0f);
                    Vector2 tPos = new Vector2(((float)i / pixelLength), ((float)j / pixelLength));
                    point_vertices[index] = new VertexPositionNormalTexture(pPos, nPos, tPos);
                    index++;
                }
            }

            // Set the data
            point_vertexBuffer.SetData<VertexPositionNormalTexture>(point_vertices);
            SPATIAL_VERTEX_SIZE = SPATIAL_TARGET_SIZE * SPATIAL_TARGET_SIZE;
        }

        // Main
        public void ExecuteRTT(float dt, GpuObject gpuObject)
        {
            // Set shader dt 
            DT.SetValue(dt);

            // Spatial Updates
            effect.CurrentTechnique = effect.Techniques["SpatialMain"];
            
            // LocationMap Update
            ExecuteSpatialRTT(ref locationMap.writeRenderTarget, ref locationMap.write);
            locationMap.SwapBuffers();
            //gpuObject.WriteToFile();

            // Indexed Updates 
            effect.CurrentTechnique = effect.Techniques["IndexedMain"];

            // Total Force Update ( collision detection )
            IndexedPass.SetValue(0);
            ExecuteIndexedRTT(ref gpuObject.totalForce.writeRenderTarget, ref gpuObject.totalForce.write);
            gpuObject.totalForce.SwapBuffers();
            //gpuObject.totalForce.WriteToFile();

            // Acceleration Update
            IndexedPass.SetValue(1);
            ExecuteIndexedRTT(ref gpuObject.acceleration.writeRenderTarget, ref gpuObject.acceleration.write);
            gpuObject.acceleration.SwapBuffers();
            //gpuObject.acceleration.WriteToFile();

            // Velocity Update
            IndexedPass.SetValue(2);
            ExecuteIndexedRTT(ref gpuObject.velocity.writeRenderTarget, ref gpuObject.velocity.write);
            gpuObject.velocity.SwapBuffers();
            //gpuObject.velocity.WriteToFile();
            
            // Rotation Update
            IndexedPass.SetValue(3);
            ExecuteIndexedRTT(ref gpuObject.rotation.writeRenderTarget, ref gpuObject.rotation.write);
            gpuObject.rotation.SwapBuffers();
            //gpuObject.rotation.WriteToFile();

            // LastPosition Update
            IndexedPass.SetValue(4);
            ExecuteIndexedRTT(ref gpuObject.lastPosition.writeRenderTarget, ref gpuObject.lastPosition.write);
            gpuObject.lastPosition.SwapBuffers();
            //gpuObject.lastPosition.WriteToFile();

            // Position Update
            IndexedPass.SetValue(5);
            ExecuteIndexedRTT(ref gpuObject.currentPosition.writeRenderTarget, ref gpuObject.currentPosition.write);
            gpuObject.currentPosition.SwapBuffers();
            //gpuObject.currentPosition.WriteToFile();

            // Dummy Update
            IndexedPass.SetValue(6);
            ExecuteIndexedRTT(ref gpuObject.dummy.writeRenderTarget, ref gpuObject.dummy.write);
            gpuObject.dummy.SwapBuffers();
            //gpuObject.dummy.WriteToFile();

            // Save texture to other fx file
            currentPositionParam.SetValue(gpuObject.currentPosition.write);
            if (initializeConstants)
            {
                initializeConstants = false;

                // Store the constants
                scaleRange.SetValue(gpuObject.currentPosition.scaleRange);
                minimumValue.SetValue(gpuObject.currentPosition.minimumValue);
            }
        }

        private void ExecuteIndexedRTT(ref RenderTarget2D renderTarget, ref Texture2D renderTargetTexture)
        {
            // Set up Graphics Device
            GraphicsDevice g = ElementalGame.graphics.GraphicsDevice;
            g.Viewport = IndexedViewport;
            g.RenderState.DepthBufferEnable = false;

            // Setup view and projection matrixes
            Matrix proj = Matrix.CreateOrthographic(1.0f, 1.0f, 1, 5);
            Matrix view = Matrix.CreateLookAt(Vector3.Backward, Vector3.Zero, new Vector3(0, 1, 1));

            effect.Parameters["ViewProj"].SetValue(view * proj);

            // Set Quad Vertex Buffer
            g.Vertices[0].SetSource(quad_vertexBuffer, 0, VertexPositionNormalTexture.SizeInBytes);

            // Set and clear RTT texture
            g.SetRenderTarget(0, renderTarget);
            g.Clear(ClearOptions.Target, Color.TransparentWhite, 1.0f, 0);

            // Draw Quad
            effect.Begin();
            foreach (EffectPass pass in effect.CurrentTechnique.Passes)
            {
                pass.Begin();
                g.VertexDeclaration = quad_vertexDeclaration;
                g.DrawPrimitives(PrimitiveType.TriangleFan, 0, 2);
                pass.End();
            }
            effect.End();

            // Gets the Rendered Texture and store it in the RenderTarget variable
            g.ResolveRenderTarget(0);

            //renderTarget.GetTexture().Save("indexedTexture.png", ImageFileFormat.Png);
            
            // Retrieve the Shadow map from our Render Target  
            //renderTargetTexture = renderTarget.GetTexture();
            //renderTargetTexture.Save("indexedTexture.png", ImageFileFormat.Png);
        }

        private void ExecuteSpatialRTT(ref RenderTarget2D renderTarget, ref Texture2D renderTargetTexture)
        {
            // Set up Graphics Device
            GraphicsDevice g = ElementalGame.graphics.GraphicsDevice;
            g.Viewport = SpatialViewport;
            g.RenderState.DepthBufferEnable = false;

            // Setup view and projection matrixes
            Matrix proj = Matrix.CreateOrthographic(0.5f, 0.5f, 0, 5);
            //Matrix view = Matrix.CreateLookAt(Vector3.Backward, Vector3.Zero, new Vector3(0, 1, 1));
            Matrix view = Matrix.CreateLookAt(new Vector3(0.75f, 0.75f, 1.0f), new Vector3(0.75f, 0.75f, 0.0f), new Vector3(0.0f, 1.0f, 0.0f));

            effect.Parameters["ViewProj"].SetValue(view * proj);
            effect.Parameters["gpuObjectScale"].SetValue(Terrain.collisionMapSize);

            // Set Points Vertex Buffer
            g.Vertices[0].SetSource(point_vertexBuffer, 0, VertexPositionNormalTexture.SizeInBytes);

            // Set and clear RTT texture
            g.SetRenderTarget(0, renderTarget);
            g.Clear(ClearOptions.Target, Color.TransparentWhite, 1.0f, 0);

            // Draw Quad
            effect.Begin();
            foreach (EffectPass pass in effect.CurrentTechnique.Passes)
            {
                pass.Begin();
                g.VertexDeclaration = point_vertexDeclaration;
                g.DrawPrimitives(PrimitiveType.PointList, 0, SPATIAL_VERTEX_SIZE);
                pass.End();
            }
            effect.End();

            // Gets the Rendered Texture and store it in the RenderTarget variable
            g.ResolveRenderTarget(0);

            //renderTarget.GetTexture().Save("gpuPhysicsTex.png", ImageFileFormat.Png);
        }
    }
}
